/*
    Project: TCGINA
    Version: V1.25
    File   : tcdebug.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcpriv.h"
#include <tchar.h>

#ifdef TC_DEBUG

#include <stdio.h>
#include <stdarg.h>

#define TCG_REGKEY_NAME \
  TEXT("SOFTWARE\\TCGINA DLL\\{D47DF546-F16B-490b-A6D8-523603A5594D}")

static HINSTANCE s_hinstDLL = NULL;
static BOOL s_fRemoteSession = FALSE;

void WINAPI TC_DbgSetHinstanceFunc(HINSTANCE hinstDLL)
{
    s_hinstDLL = hinstDLL;
    s_fRemoteSession = GetSystemMetrics(SM_REMOTESESSION) != 0;
}

static void WINAPI TC_DbgPrintModuleInfo(HINSTANCE hinst)
{
    CHAR   aszFileName[MAX_PATH];
    DWORD  dwFileVersionInfoSize;
    DWORD  dwResult;
    HANDLE h;
    WIN32_FIND_DATAA fd;

    if (GetModuleFileNameA(hinst, aszFileName, MAX_PATH) == 0) return;
    if (aszFileName[0] == '\\' && aszFileName[1] == '?' &&
      aszFileName[2] == '?' && aszFileName[3] == '\\') {
        memmove(aszFileName, aszFileName + 4, strlen(aszFileName) - 3);
    }

    TC_DbgPrintfFuncA("%s", aszFileName);

    dwFileVersionInfoSize = GetFileVersionInfoSizeA(aszFileName, &dwResult);
    if (dwFileVersionInfoSize) {
        PVOID pBlock = LocalAlloc(LMEM_FIXED, dwFileVersionInfoSize);
        if (pBlock) {
            if (GetFileVersionInfoA(aszFileName, 0, dwFileVersionInfoSize,
              pBlock)) {
                VS_FIXEDFILEINFO *ffi;
                UINT uLen;
                if (VerQueryValue(pBlock, TEXT("\\"), (LPVOID *)&ffi, &uLen) &&
                  ffi->dwSignature == 0xfeef04bd) {
                    TC_DbgPrintfFuncA(" V%d.%d.%d.%d",
                      ffi->dwFileVersionMS >> 16,
                      ffi->dwFileVersionMS & 0xffff,
                      ffi->dwFileVersionLS >> 16,
                      ffi->dwFileVersionLS & 0xffff);
                }
            }
            LocalFree(pBlock);
        }
    }

    h = FindFirstFileA(aszFileName, &fd);
    if (h != INVALID_HANDLE_VALUE) {
        SYSTEMTIME st, lt;
        if (FileTimeToSystemTime(&fd.ftLastWriteTime, &st)) {
            if (SystemTimeToTzSpecificLocalTime(NULL, &st, &lt)) {
                TC_DbgPrintfFuncA(" %04d-%02d-%02d %02d:%02d:%02d",
                    (int)lt.wYear,
                    (int)lt.wMonth,
                    (int)lt.wDay,
                    (int)lt.wHour,
                    (int)lt.wMinute,
                    (int)lt.wSecond);
            }
        }
        FindClose(h);
    }

    TC_DbgPrintfFuncA("\n");
}

void WINAPI TC_DbgPutsFuncA(LPCSTR psz)
{
    static BOOL  fDbgInsertTimeStamp = TRUE;
    static TCHAR szFileName[MAX_PATH];
    FILE *fp;

    if (fDbgInsertTimeStamp) {

        SYSTEMTIME lt;

        fDbgInsertTimeStamp = FALSE;

        *szFileName = TEXT('\0');
        GetEnvironmentVariable(TEXT("SystemDrive"), szFileName, MAX_PATH);
        if (*szFileName == TEXT('\0')) {
            GetEnvironmentVariable(TEXT("SystemRoot"), szFileName, MAX_PATH);
            if (*szFileName == TEXT('\0')) {
                TCHAR chDrive;
                _tcscpy(szFileName, TEXT("C:"));
                for (chDrive = TEXT('A'); chDrive <= TEXT('Z'); chDrive++) {
                    if (TC_GetDriveType(chDrive) == DRIVE_FIXED) {
                        *szFileName = chDrive;
                        break;
                    }
                }
            }
        }
        switch(szFileName[_tcslen(szFileName) - 1]) {
        case '/':
        case '\\':
            break;
        default:
            TC_AppendString(szFileName, MAX_PATH, TEXT("\\"));
        }
        TC_AppendString(szFileName, MAX_PATH, TEXT("TC.LOG"));
        TC_GetRegistryString(HKEY_LOCAL_MACHINE, TCG_REGKEY_NAME,
          TEXT("LogFileName"), szFileName, MAX_PATH);

        GetLocalTime(&lt);
        TC_DbgPrintfFuncA("\n%04d-%02d-%02d %02d:%02d:%02d\n",
            (int)lt.wYear,
            (int)lt.wMonth,
            (int)lt.wDay,
            (int)lt.wHour,
            (int)lt.wMinute,
            (int)lt.wSecond);
        TC_DbgPrintModuleInfo(0);
        if (s_hinstDLL) TC_DbgPrintModuleInfo(s_hinstDLL);
    }

    fp = _tfopen(szFileName, TEXT("a"));
    if (fp) {
        if (s_fRemoteSession) {
            static DWORD dwCol = 0;
            for (; *psz; psz++) {
                if (dwCol == 0) fputs("[R] ", fp);
                fputc(*psz, fp);
                if (*psz != '\n') {
                    dwCol++;
                } else {
                    dwCol = 0;
                }
            }
        } else {
            fputs(psz, fp);
        }
        fclose(fp);
    }
}

void WINAPI TC_DbgPutsFuncW(LPCWSTR pwsz)
{
    CHAR aszBuffer[512];

    if (TC_ConvertUnicodeToAnsi(aszBuffer, sizeof(aszBuffer), pwsz)) {
        TC_DbgPutsFuncA(aszBuffer);
    }
}

#define TC_DBG_BUFFER_SIZE 512

void TC_DbgPrintfFuncA(LPCSTR paszFormat, ...)
{
    CHAR aszBuffer[TC_DBG_BUFFER_SIZE];
    va_list ap;

    va_start(ap, paszFormat);
    _vsnprintf(aszBuffer, TC_DBG_BUFFER_SIZE, paszFormat, ap);
    aszBuffer[TC_DBG_BUFFER_SIZE - 1] = '\0';
    va_end(ap);

    TC_DbgPutsFuncA(aszBuffer);
}

void TC_DbgPrintfFuncW(LPWSTR pwszFormat, ...)
{
    WCHAR wszBuffer[TC_DBG_BUFFER_SIZE];
    va_list ap;

    va_start(ap, pwszFormat);
    _vsnwprintf(wszBuffer, TC_DBG_BUFFER_SIZE, pwszFormat, ap);
    wszBuffer[TC_DBG_BUFFER_SIZE - 1] = TEXT('\0');
    va_end(ap);

    TC_DbgPutsFuncW(wszBuffer);
}

void WINAPI TC_AssertFunc(BOOL fAssertion, LPCSTR pszFileName, DWORD dwLine)
{
    if (!fAssertion) {
        TC_DbgPrintfA("Assertion failed in \"%s\" at line %u\n",
          pszFileName, dwLine);
    }
}

typedef struct {
    LPCSTR pszFileName;
    DWORD  dwLine;
    SIZE_T uSize;
    DWORD  dwSignature1;
} TC_ALLOC_DATA, *PTC_ALLOC_DATA;

#define TC_ALLOC_SIGNATURE1 0xe7afd6b8
#define TC_ALLOC_SIGNATURE2 0x2e9f0ab5

#define TC_InternalAlloc(x) LocalAlloc(LMEM_FIXED,x)
#define TC_InternalFree(p)  LocalFree(p)

PVOID WINAPI TC_AllocFunc(SIZE_T uBytes, LPCSTR pszFileName, DWORD dwLine)
{
    PTC_ALLOC_DATA pad;
    LPVOID p;

    if (uBytes == 0) {
        TC_DbgPrintfA(
          "TC_Alloc() called with zero size in \"%s\" at line %u\n",
          pszFileName, dwLine);
        return NULL;
    }

    pad = TC_InternalAlloc(uBytes + (sizeof(TC_ALLOC_DATA) + sizeof(DWORD)));
    if (pad == NULL) {
        TC_DbgPrintfA("TC_Alloc() failed in \"%s\" at line %u\n",
          pszFileName, dwLine);
        return NULL;
    }
    pad->pszFileName = pszFileName;
    pad->dwLine = dwLine;
    pad->uSize = uBytes;
    pad->dwSignature1 = TC_ALLOC_SIGNATURE1;
    p = ((LPBYTE)pad) + sizeof(TC_ALLOC_DATA);
    memset(p, '\xaa', uBytes);
    *(DWORD*)(((LPBYTE)p)+uBytes) = TC_ALLOC_SIGNATURE2;

    return p;
}

PVOID WINAPI TC_CallocFunc(SIZE_T uBytes, LPCSTR pszFileName, DWORD dwLine)
{
    LPVOID p;

    p = TC_AllocFunc(uBytes, pszFileName, dwLine);
    if (p) TC_ZeroMemory(p, uBytes);

    return p;
}

void WINAPI TC_FreeFunc(PVOID p, LPCSTR pszFileName, DWORD dwLine)
{
    PTC_ALLOC_DATA pad;

    if (p == NULL) {
        TC_DbgPrintfA("TC_Free() called with NULL pointer in "
          "\"%s\" at line %u\n", pszFileName, dwLine);
        return;
    }
    pad = (PTC_ALLOC_DATA)(((LPBYTE)p) - sizeof(TC_ALLOC_DATA));
    if (pad->dwSignature1 != TC_ALLOC_SIGNATURE1) {
        TC_DbgPrintfA("Invalid pointer or buffer underrun detected in "
         "\"%s\n at line %u\n", pszFileName, dwLine);
    } else {
        switch(*(DWORD*)(((LPBYTE)p)+pad->uSize)) {
        case TC_ALLOC_SIGNATURE2:
            memset(p, '\x55', pad->uSize);
            *(DWORD*)(((LPBYTE)p)+pad->uSize) = (DWORD)~TC_ALLOC_SIGNATURE2;
            break;

        case ~TC_ALLOC_SIGNATURE2:
            TC_DbgPrintfA("TC_Free() called for already freed block in "
              "\"%s\" at line %u\n", pszFileName, dwLine);
            TC_DbgPrintfA("Block of %u bytes was allocated in \"%s\" at "
              "line %u\n", pad->uSize, pad->pszFileName, pad->dwLine);
            break;

        default:
            TC_DbgPrintfA("Buffer overrun detected in \"%s\" at line %u\n",
              pszFileName, dwLine);
            TC_DbgPrintfA("Block of %u bytes was allocated in \"%s\" at "
              "line %u\n", pad->uSize, pad->pszFileName, pad->dwLine);
        }
    }

    TC_InternalFree((LPVOID)pad);
}

PVOID WINAPI TC_ConvertLocalAllocBlockFunc(PVOID p, SIZE_T uBytes,
  LPCSTR pszFileName, DWORD dwLine)
{
    /*
        Convert a data block which has been allocated with LocalAlloc()
        into a data block which has been allocated with TC_Alloc()
    */

    PVOID p2;

    TC_Assert(p);

    if (p == 0) return 0;

    p2 = TC_AllocFunc(uBytes, pszFileName, dwLine);
    if (p2) memcpy(p2, p, uBytes);
    LocalFree(p);

    return p2;
}

#endif
