/*
    Project: TCGINA
    Version: V1.25
    File   : user.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <winioctl.h>
#include <tchar.h>
#include <tc.h>
#include <stdlib.h>
#include "setup.h"

typedef struct {
    ULONG  ReparseTag;
    USHORT ReparseDataLength;
    USHORT Reserved;
    USHORT SubstituteNameOffset;
    USHORT SubstituteNameLength;
    USHORT PrintNameOffset;
    USHORT PrintNameLength;
    WCHAR  PathBuffer[1];
} REPARSE_DATA_BUFFER, *PREPARSE_DATA_BUFFER;

static BOOL GetSubstituteName(LPCTSTR pszPathName, LPTSTR pszDest,
  DWORD nSize)
{
    HANDLE hFile;
    BOOL   fResult;
    BYTE   rgbBuffer[sizeof(REPARSE_GUID_DATA_BUFFER) +
      MAXIMUM_REPARSE_DATA_BUFFER_SIZE];
    DWORD  dwSize;

    hFile = CreateFile(pszPathName, 0,
      FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
      FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
        return FALSE;
    }
    fResult = FALSE;
    if (DeviceIoControl(hFile, FSCTL_GET_REPARSE_POINT, NULL, 0,
      rgbBuffer, sizeof(rgbBuffer), &dwSize, NULL)) {
        PREPARSE_DATA_BUFFER prdb = (PREPARSE_DATA_BUFFER) rgbBuffer;
        if (prdb->ReparseTag == IO_REPARSE_TAG_MOUNT_POINT) {
            LPWSTR pwszSubstituteName = (LPWSTR)
              ( ((PBYTE)prdb->PathBuffer) + prdb->SubstituteNameOffset );
            pwszSubstituteName[prdb->SubstituteNameLength/sizeof(WCHAR)] =
              L'\0';
            if (TC_CopyStringFromUnicode(pszDest, nSize,
              pwszSubstituteName) && *pszDest) {
                fResult = TRUE;
            }
        }
    }
    CloseHandle(hFile);

    return fResult;
}

static BOOL IsTrueCryptMountPoint(PUSER_INFO pui, LPCTSTR pszPathName)
{
    TCHAR szFileName[MAX_PATH];

    UNREFERENCED_PARAMETER(pui);

    if (!TC_PathIsMountPoint(pszPathName)) {
        return FALSE;
    }

    if (GetSubstituteName(pszPathName, szFileName, MAX_PATH)) {
        BYTE   rgbUniqueId[MAX_PATH];
        DWORD  cbSize;
        LPTSTR pszTemp;

        pszTemp = _tcschr(szFileName, TEXT('}'));
        if (pszTemp) {
            if (pszTemp[1] == TEXT('\\')) {
                pszTemp[1] = TEXT('\0');
            }
        }
        cbSize = sizeof(rgbUniqueId);
        memset(rgbUniqueId, '\0', sizeof(rgbUniqueId));
        if (RegGetValueData(HKEY_LOCAL_MACHINE, TEXT("SYSTEM\\MountedDevices"),
          szFileName, NULL, rgbUniqueId, &cbSize)) {
            if (memcmp(rgbUniqueId, "TrueCrypt", 9) == 0) {
                return TRUE;
            } else {
                return FALSE;
            }
        }
    }

    /*
        A mount point is considered as encrypted location if a
        .TCUSER file exists at its root.
    */
    TC_CopyString(szFileName, MAX_PATH, pszPathName);
    TC_AppendString(szFileName, MAX_PATH, TEXT("\\*.TCUSER"));
    if (FileExists(szFileName)) {
        return TRUE;
    }

    return FALSE;
}

static BOOL ProfileIsEncrypted(PUSER_INFO pui, PUSER_INFO_ITEM puii)
{
    TCHAR szFileName[MAX_PATH];

    TC_CopyString(szFileName, MAX_PATH, puii->szProfileImagePath);
    TC_AppendString(szFileName, MAX_PATH, TEXT("\\NTUSER.DAT"));

    if (!FileExists(szFileName)) {
        return TRUE;
    }

    if (_istalpha(*szFileName) && szFileName[1] == ':') {
        int nDrive = _totupper(*szFileName) - TEXT('A');
        if (*pui->rgszTCFileName[nDrive]) {
            return TRUE;
        }
    }

    return IsTrueCryptMountPoint(pui, puii->szProfileImagePath);
}

static BOOL EncryptedProfileIsUninstallable(PUSER_INFO pui,
  PUSER_INFO_ITEM puii)
{
    TCHAR szKeyName[MAX_PATH];
    TCHAR szSID[MAX_PATH];
    TCHAR szProfileImagePath[MAX_PATH];
    TCHAR szFileName[MAX_PATH];

    UNREFERENCED_PARAMETER(pui);

    TC_CopyString(szKeyName, MAX_PATH, TCGINA_KEY);
    TC_AppendString(szKeyName, MAX_PATH, TEXT("\\"));
    TC_AppendString(szKeyName, MAX_PATH, puii->szUserName);

    *szSID = TEXT('\0');
    TC_GetRegistryString(HKEY_LOCAL_MACHINE, szKeyName, TEXT("SID"),
      szSID, MAX_PATH);
    *szProfileImagePath = TEXT('\0');
    TC_GetRegistryString(HKEY_LOCAL_MACHINE, szKeyName, TEXT("ProfileImagePath"),
      szProfileImagePath, MAX_PATH);

    if (*szSID == TEXT('\0') || *szProfileImagePath == TEXT('\0') ||
      _tcscmp(szSID, puii->szSID) != 0) {
        return FALSE;
    }

    TC_CopyString(szFileName, MAX_PATH, szProfileImagePath);
    TC_AppendString(szFileName, MAX_PATH, TEXT("\\NTUSER.DAT"));

    if (!FileExists(szFileName)) {
        return FALSE;
    }
    if (IsTrueCryptMountPoint(pui, szProfileImagePath)) {
        return FALSE;
    }

    return TRUE;
}

BOOL UpdateTCFileNames(PUSER_INFO pui)
{
    TC_HANDLE hTC;
    TCHAR     szDrive[4];
    DWORD     dwMountedDrives;
    int       i;

    if (pui == NULL) {
        return FALSE;
    }

    hTC = TC_Open(0);
    if (hTC == NULL) {
        return FALSE;
    }

    pui->fTCFileNamesValid = FALSE;
    memset(pui->rgszTCFileName, '\0', sizeof(pui->rgszTCFileName));
    szDrive[1] = TEXT(':');
    szDrive[2] = TEXT('\\');
    szDrive[3] = TEXT('\0');

    if (!TC_GetMountedDrives(hTC, &dwMountedDrives, TCMD_DISMOUNTABLE)) {
        dwMountedDrives = (DWORD)-1;
    }

    for (i = 0; i < 26; i++) {
        if ((dwMountedDrives & (1<<i)) == 0) continue;
        szDrive[0] = (TCHAR)(TEXT('A') + i);
        if (!TC_GetVolumeFileName(hTC, szDrive, pui->rgszTCFileName[i],
          MAX_PATH)) {
            TC_Close(hTC);
            return FALSE;
        }
    }
    TC_Close(hTC);
    pui->fTCFileNamesValid = TRUE;

    return TRUE;
}

static int __cdecl CompareUserInfoItem(const void *p1, const void *p2)
{
    PUSER_INFO_ITEM uii1 = (PUSER_INFO_ITEM) p1;
    PUSER_INFO_ITEM uii2 = (PUSER_INFO_ITEM) p2;
    int iResult = CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
      uii1->szUserName, -1, uii2->szUserName, -1);
    if (iResult) iResult -= 2;

    return iResult;
}

PUSER_INFO GetUserInfo(void)
{
    PUSER_INFO pui;
    HKEY       hKey;
    DWORD      dwSize;
    TCHAR      szLogonUserName[MAX_PATH];

    pui = calloc(1, sizeof(USER_INFO));
    if (pui == NULL) {
        return NULL;
    }
    UpdateTCFileNames(pui);

    dwSize = MAX_PATH;
    if (!GetUserName(szLogonUserName, &dwSize)) {
        *szLogonUserName = TEXT('\0');
    }

    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, PROFILELIST_KEY, 0, KEY_READ,
      &hKey) == ERROR_SUCCESS) {
        FILETIME       ftLastWriteTime;
        TCHAR          szKeyName[MAX_PATH];
        DWORD          dwIndex;
        USER_INFO_ITEM uii;

        dwIndex = 0;
        dwSize = MAX_PATH;
        while (RegEnumKeyEx(hKey, dwIndex, szKeyName, &dwSize,
          NULL, NULL, NULL, &ftLastWriteTime) == ERROR_SUCCESS) {
            DWORD   dwHyphenCount = 0;
            LPTSTR  pszTemp;
            LPCTSTR pszUserName;

            pszUserName = TC_GetUserName(szKeyName);
            if (pszUserName) {
                for (pszTemp = szKeyName; *pszTemp; pszTemp++) {
                    if (*pszTemp == TEXT('-')) dwHyphenCount++;
                }
                if (*pszUserName && dwHyphenCount >= 7) {
                    PUSER_INFO_ITEM pNewItem;
                    memset(&uii, '\0', sizeof(uii));
                    TC_CopyString(uii.szSID, MAX_PATH, szKeyName);
                    TC_GetRegistryString(hKey, szKeyName,
                      TEXT("ProfileImagePath"),
                      uii.szProfileImagePath, MAX_PATH);
                    TC_CopyString(uii.szUserName, MAX_PATH,
                        pszUserName);
                    if (_tcsicmp(uii.szUserName, szLogonUserName) != 0) {
                        if (ProfileIsEncrypted(pui, &uii)) {
                            uii.fEncryptedProfile = TRUE;
                            pui->dwEncryptedUsers++;
                            if (EncryptedProfileIsUninstallable(pui, &uii)) {
                                uii.fUninstallable = TRUE;
                                pui->dwUninstallableUsers++;
                            }
                        } else {
                            pui->dwUnencryptedUsers++;
                        }
                        pNewItem = realloc(pui->pItem,
                          (pui->dwUserCount+1) * sizeof(uii));
                        if (pNewItem == NULL) {
                            RegCloseKey(hKey);
                            free(pui->pItem);
                            free(pui);
                            return NULL;
                        }
                        pui->pItem = pNewItem;
                        pui->pItem[pui->dwUserCount++] = uii;
                    }
                }
                TC_Free(pszUserName);
            }
            dwSize = MAX_PATH;
            dwIndex++;
        }
        RegCloseKey(hKey);
    }

    if (pui->dwUserCount > 1) {
        qsort(pui->pItem, pui->dwUserCount, sizeof(USER_INFO_ITEM),
          CompareUserInfoItem);
    }

    return pui;
}

void FreeUserInfo(PUSER_INFO pui)
{
    if (pui) {
        free(pui->pItem);
        free(pui);
    }
}